/*
 * Simulator of MCS51 (uc51r.cc)
 *
 * Copyright (C) 1997,97 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of MCS51 simulator: s51.

S51 is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

S51 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with S51; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>

#include "uc51rcl.h"


/*
 * Making an 8052 CPU object
 *____________________________________________________________________________
 *
 */

t_uc51r::t_uc51r(int Itype, int Itech,
		 //FILE *in, FILE *out,
		 FILE *ser_in, FILE *ser_out,
		 bool verbose, class cl_sim *asim):
  t_uc52(Itype, Itech, ser_in, ser_out, verbose, asim)
{
  int i;

  for (i= 0; i < ERAM_SIZE; i++)
    ERAM[i]= 0;
  clock_out= 0;
}


/*
 * Resetting of the microcontroller
 *____________________________________________________________________________
 *
 * Original method is extended with handling of WDT.
 *
 */

void
t_uc51r::reset(void)
{
  t_uc52::reset();
  WDT= -1; // Disable WDT
  wdtrst= 0;
}


/*
 * Copying ERAM to XRAM and vice versa
 *____________________________________________________________________________
 *
 * This two methods are used by command interpreter to make ERAM and
 * beginning of XRAM to be equivalent.
 *
 */

void
t_uc51r::eram2xram(void)
{
  int i;

  for (i= 0; i < ERAM_SIZE; i++)
    set_mem(MEM_XRAM, i, ERAM[i]);
}

void
t_uc51r::xram2eram(void)
{
  int i;

  for (i= 0; i < ERAM_SIZE; i++)
    ERAM[i]= get_mem(MEM_XRAM, i);
}


/*
 * Processing write operation of SFR
 *____________________________________________________________________________
 *
 * Inherited method is extended with WDT handling.
 *
 */

void
t_uc51r::proc_write(uchar *addr)
{
  t_uc52::proc_write(addr);
  // Handling WDT
  if (addr == &SFR[WDTRST])
    {
      if ((wdtrst == 0x1e) &&
	  (*addr == 0xe1))
	WDT= 0;
      wdtrst= *addr;
    }
}


/*
 * Simulating timers
 *____________________________________________________________________________
 *
 * Calling inherited method to simulate timer #0 and #1 and then 
 * simulating timer #2.
 *
 */

int
t_uc51r::do_timers(int cycles)
{
  int res;

  if ((res= t_uc51::do_timers(cycles)) != resGO)
    return(res);
  return(do_timer2(cycles));
}


/*
 * Simulating timer 2
 *____________________________________________________________________________
 *
 * It is something wrong: T2MOD is not implemented in 52?!
 */

int
t_uc51r::do_timer2(int cycles)
{
  bool nocount= FALSE;

  exf2it->activate();
  if (!(SFR[T2CON] & bmTR2))
    /* Timer OFF */
    return(resGO);

  if (SFR[T2MOD] & bmT2OE)
    return(do_t2_clockout(cycles));

  if (SFR[T2CON] & (bmRCLK | bmTCLK))
    return(do_t2_baud(cycles));

  /* Determining nr of input clocks */
  if (!(SFR[T2CON] & bmTR2))
    nocount= TRUE; // Timer OFF
  else
    if (SFR[T2CON] & bmC_T2)
      {
	// Counter mode, falling edge on P1.0 (T2)
	if ((prev_p1 & bmT2) &&
	    !(SFR[P1] & port_pins[1] & bmT2))
	  cycles= 1;
	else
	  nocount= TRUE;
      }
  /* Counting */
  while (cycles--)
    {
      if (SFR[T2CON] & bmCP_RL2)
	do_t2_capture(&cycles, nocount);
      else
	{
	  int overflow;
	  overflow= 0;
	  /* Auto-Relode mode */
	  if (SFR[T2MOD] & bmDCEN)
	    {
	      /* DCEN= 1 */
	      exf2it->deactivate();
	      if (nocount)
		cycles= 0;
	      else
		{
		  if (SFR[P1] & port_pins[1] & bmT2EX)
		    {
		      // UP
		      if (!++SFR[TL2])
			if (!++SFR[TH2])
			  {
			    overflow++;
			    SFR[TH2]= SFR[RCAP2H];
			    SFR[TL2]= SFR[RCAP2L];
			    SFR[T2CON]|= bmTF2;
			  }
		    }
		  else
		    {
		      // DOWN
		      SFR[TL2]--;
		      if (SFR[TL2] == 0xff)
			  SFR[TH2]--;
		      if (SFR[TH2] == SFR[RCAP2H] &&
			  SFR[TL2] == SFR[RCAP2L])
			{
			  overflow++;
			  SFR[TH2]= SFR[TL2]= 0xff;
			  SFR[T2CON]|= bmTF2;
			}
		    }
		  while (overflow--)
		      SFR[P1]^= bmEXF2;
		}
	    }
	  else
	    /* DCEN= 0 */
	    do_t2_reload(&cycles, nocount);
	}
    }// while cycles

  return(resGO);
}


/*
 * Clock out mode of Timer #2
 *____________________________________________________________________________
 *
 */

int
t_uc51r::do_t2_clockout(int cycles)
{
  /* Programmable Clock Out Mode */
  if ((prev_p1 & bmT2EX) &&
      !(SFR[P1] & port_pins[1] & bmT2EX) &&
      (SFR[T2CON] & bmEXEN2))
    SFR[T2CON]|= bmEXF2;
  if (SFR[T2CON] & bmCP_RL2)
    return(resGO);
  if (SFR[T2CON] & bmC_T2)
    {
      if ((prev_p1 & bmT2) &&
	  !(SFR[P1] & port_pins[1] & bmT2))
	cycles= 1;
      else
	cycles= 0;
    }
  else
    cycles*= 6;
  if (SFR[T2CON] & bmTR2)
    while (cycles--)
      {
	if (!++SFR[TL2])
	  if (!++SFR[TH2])
	    {
	      SFR[TH2]= SFR[RCAP2H];
	      SFR[TL2]= SFR[RCAP2L];
	      clock_out++;
	      if (!(SFR[T2CON] & bmC_T2))
		if (clock_out & 1)
		  SFR[P1]|= bmT2;
		else
		  SFR[P1]&= ~bmT2;
	    }
      }
  return(resGO);
}


/*
 *
 */

int
t_uc51r::serial_bit_cnt(void)
{
  int mode, divby;
  int *tr_src, *rec_src;

  mode= SFR[SCON] >> 6;
  switch (mode)
    {
    case 0:
      divby  = 12;
      tr_src = &s_tr_tick;
      rec_src= &s_rec_tick;
      break;
    case 1:
    case 3:
      divby  = (SFR[PCON]&bmSMOD)?16:32;
      tr_src = (SFR[T2CON]&bmTCLK)?(&s_tr_t2):(&s_tr_t1);
      rec_src= (SFR[T2CON]&bmTCLK)?(&s_rec_t2):(&s_rec_t1);
      break;
    case 2:
      divby  = (SFR[PCON]&bmSMOD)?16:32;
      tr_src = &s_tr_tick;
      rec_src= &s_rec_tick;
      break;
    }
  if (s_sending)
    {
      while (*tr_src >= divby)
	{
	  (*tr_src)-= divby;
	  s_tr_bit++;
	}
    }
  if (s_receiving)
    {
      while (*rec_src >= divby)
	{
	  (*rec_src)-= divby;
	  s_rec_bit++;
	}
    }
}


/*
 * Handling WDT
 *____________________________________________________________________________
 *
 */

int
t_uc51r::do_wdt(int cycles)
{
  if (WDT >= 0)
    {
      WDT+= cycles;
      if (WDT & ~(0x3fff))
	{
	  reset();
	  return(resWDTRESET);
	}
    }
  return(resGO);
}


/*
 * 0xe0 1 24 MOVX A,@DPTR
 *____________________________________________________________________________
 *
 */

int
t_uc51r::inst_movx_a_$dptr(uchar code)
{
  if ((SFR[AUXR] & bmEXTRAM) ||
      SFR[DPH])
    SFR[event_at.ws= ACC]= read_mem(MEM_XRAM,
				    event_at.rx= SFR[DPH]*256+SFR[DPL]);
  else
    SFR[event_at.ws= ACC]= ERAM[event_at.rx= SFR[DPL]];
  tick(1);
  return(resGO);
}


/*
 * 0xe2-0xe3 1 24 MOVX A,@Ri
 *____________________________________________________________________________
 *
 */

int
t_uc51r::inst_movx_a_$ri(uchar code)
{
  uchar *addr;
  int res;

  addr= get_indirect(*(get_reg(code & 0x01)), &res);
  if (SFR[AUXR] & bmEXTRAM)
    SFR[event_at.ws= ACC]=
      read_mem(MEM_XRAM, event_at.rx= (SFR[P2]&port_pins[2])*256+*addr);
  else
    SFR[event_at.ws= ACC]= ERAM[event_at.rx= *addr];
  tick(1);
  return(res);
}


/*
 * 0xf0 1 24 MOVX @DPTR,A
 *____________________________________________________________________________
 *
 */

int
t_uc51r::inst_movx_$dptr_a(uchar code)
{
  if ((SFR[AUXR] & bmEXTRAM) ||
      SFR[DPH])
    write_mem(MEM_XRAM,
	      event_at.wx= SFR[DPH]*256+SFR[DPL],
	      SFR[event_at.rs= ACC]);
  else
    ERAM[event_at.wx= SFR[DPL]]= SFR[event_at.rs= ACC];
  return(resGO);
}


/*
 * 0xf2-0xf3 1 24 MOVX @Ri,A
 *____________________________________________________________________________
 *
 */

int
t_uc51r::inst_movx_$ri_a(uchar code)
{
  uchar *addr;
  int res;

  addr= get_indirect(event_at.wi= *(get_reg(code & 0x01)), &res);
  if (SFR[AUXR] & bmEXTRAM)
    write_mem(MEM_XRAM,
	      event_at.wx= (SFR[P2] & port_pins[2])*256 + *addr,
	      SFR[ACC]);
  else
    ERAM[event_at.wx= *addr]= SFR[ACC];
  tick(1);
  return(res);
}


/* End of uc51r.cc */
